@@ -24,6 +24,7 @@ from contextlib import contextmanager
 from coconut.exceptions import (
     CoconutInternalException,
     CoconutDeferredSyntaxError,
+    internal_assert,
 )
 from coconut.constants import (
     match_temp_var,
@@ -32,7 +33,9 @@ from coconut.constants import (
     closeindent,
     match_check_var,
     const_vars,
+    sentinel_var,
 )
+from coconut.compiler.util import paren_join
 
 #-----------------------------------------------------------------------------------------------------------------------
 # UTILITIES:
@@ -72,6 +75,9 @@ class Matcher(object):
         "series": lambda self: self.match_sequence,
         "rseries": lambda self: self.match_rsequence,
         "mseries": lambda self: self.match_msequence,
+        "string": lambda self: self.match_string,
+        "rstring": lambda self: self.match_rstring,
+        "mstring": lambda self: self.match_mstring,
         "const": lambda self: self.match_const,
         "var": lambda self: self.match_var,
         "set": lambda self: self.match_set,
@@ -90,6 +96,7 @@ class Matcher(object):
         "var_index",
         "others",
         "guards",
+        "use_sentinel",
     )
 
     def __init__(self, loc, checkdefs=None, names=None, var_index=0):
@@ -107,6 +114,7 @@ class Matcher(object):
         self.var_index = var_index
         self.others = []
         self.guards = []
+        self.use_sentinel = False
 
     def duplicate(self):
         """Duplicates the matcher to others."""
@@ -188,7 +196,7 @@ class Matcher(object):
         self.set_position(self.position - by)
 
     @contextmanager
-    def incremented(self, by=1):
+    def down_a_level(self, by=1):
         """Increment then decrement."""
         self.increment(by)
         try:
@@ -213,8 +221,8 @@ class Matcher(object):
 
     def match_all_in(self, matches, item):
         """Matches all matches to elements of item."""
-        for x in range(len(matches)):
-            self.match(matches[x], item + "[" + str(x) + "]")
+        for i, match in enumerate(matches):
+            self.match(match, item + "[" + str(i) + "]")
 
     def check_len_in(self, min_len, max_len, item):
         """Checks that the length of item is in range(min_len, max_len+1)."""
@@ -228,13 +236,13 @@ class Matcher(object):
         else:
             self.add_check(str(min_len) + " <= _coconut.len(" + item + ") <= " + str(max_len))
 
-    def match_function(self, args, kwargs, match_args=[], star_arg=None, kwd_args=[], dubstar_arg=None):
+    def match_function(self, args, kwargs, match_args=(), star_arg=None, kwd_args=(), dubstar_arg=None):
         """Matches a pattern-matching function."""
         self.match_in_args_kwargs(match_args, args, kwargs, allow_star_args=star_arg is not None)
         if star_arg is not None:
             self.match(star_arg, args + "[" + str(len(match_args)) + ":]")
         self.match_in_kwargs(kwd_args, kwargs)
-        with self.incremented():
+        with self.down_a_level():
             if dubstar_arg is None:
                 self.add_check("not " + kwargs)
             else:
@@ -244,71 +252,75 @@ class Matcher(object):
         """Matches against args or kwargs."""
         req_len = 0
         arg_checks = {}
-        to_match = []  # [(increment, match, against)]
-        for x in range(len(match_args)):
-            if isinstance(match_args[x], tuple):
-                (match, default) = match_args[x]
+        to_match = []  # [(move_down, match, against)]
+        for i, arg in enumerate(match_args):
+            if isinstance(arg, tuple):
+                (match, default) = arg
             else:
-                match, default = match_args[x], None
+                match, default = arg, None
             names = get_match_names(match)
             if default is None:
                 if not names:
-                    req_len = x + 1
-                    to_match.append((False, match, args + "[" + str(x) + "]"))
+                    req_len = i + 1
+                    to_match.append((False, match, args + "[" + str(i) + "]"))
                 else:
-                    arg_checks[x] = (
-                        # if x < req_len
+                    arg_checks[i] = (
+                        # if i < req_len
                         " and ".join('"' + name + '" not in ' + kwargs for name in names),
-                        # if x >= req_len
-                        "_coconut.sum((_coconut.len(" + args + ") > " + str(x) + ", "
+                        # if i >= req_len
+                        "_coconut.sum((_coconut.len(" + args + ") > " + str(i) + ", "
                         + ", ".join('"' + name + '" in ' + kwargs for name in names)
                         + ")) == 1",
                     )
                     tempvar = self.get_temp_var()
-                    self.add_def(tempvar + " = "
-                                 + args + "[" + str(x) + "] if _coconut.len(" + args + ") > " + str(x) + " else "
-                                 + "".join(kwargs + '.pop("' + name + '") if "' + name + '" in ' + kwargs + " else "
-                                           for name in names[:-1])
-                                 + kwargs + '.pop("' + names[-1] + '")'
-                                 )
+                    self.add_def(
+                        tempvar + " = "
+                        + args + "[" + str(i) + "] if _coconut.len(" + args + ") > " + str(i) + " else "
+                        + "".join(kwargs + '.pop("' + name + '") if "' + name + '" in ' + kwargs + " else "
+                                  for name in names[:-1])
+                        + kwargs + '.pop("' + names[-1] + '")',
+                    )
                     to_match.append((True, match, tempvar))
             else:
                 if not names:
                     tempvar = self.get_temp_var()
-                    self.add_def(tempvar + " = " + args + "[" + str(x) + "] if _coconut.len(" + args + ") > " + str(x) + " else " + default)
+                    self.add_def(tempvar + " = " + args + "[" + str(i) + "] if _coconut.len(" + args + ") > " + str(i) + " else " + default)
                     to_match.append((True, match, tempvar))
                 else:
-                    arg_checks[x] = (
-                        # if x < req_len
+                    arg_checks[i] = (
+                        # if i < req_len
                         None,
-                        # if x >= req_len
-                        "_coconut.sum((_coconut.len(" + args + ") > " + str(x) + ", "
+                        # if i >= req_len
+                        "_coconut.sum((_coconut.len(" + args + ") > " + str(i) + ", "
                         + ", ".join('"' + name + '" in ' + kwargs for name in names)
                         + ")) <= 1",
                     )
                     tempvar = self.get_temp_var()
-                    self.add_def(tempvar + " = "
-                                 + args + "[" + str(x) + "] if _coconut.len(" + args + ") > " + str(x) + " else "
-                                 + "".join(kwargs + '.pop("' + name + '") if "' + name + '" in ' + kwargs + " else "
-                                           for name in names)
-                                 + default
-                                 )
+                    self.add_def(
+                        tempvar + " = "
+                        + args + "[" + str(i) + "] if _coconut.len(" + args + ") > " + str(i) + " else "
+                        + "".join(
+                            kwargs + '.pop("' + name + '") if "' + name + '" in ' + kwargs + " else "
+                            for name in names
+                        )
+                        + default,
+                    )
                     to_match.append((True, match, tempvar))
 
         max_len = None if allow_star_args else len(match_args)
         self.check_len_in(req_len, max_len, args)
-        for x in arg_checks:
-            lt_check, ge_check = arg_checks[x]
-            if x < req_len:
+        for i in sorted(arg_checks):
+            lt_check, ge_check = arg_checks[i]
+            if i < req_len:
                 if lt_check is not None:
                     self.add_check(lt_check)
             else:
                 if ge_check is not None:
                     self.add_check(ge_check)
 
-        for increment, match, against in to_match:
-            if increment:
-                with self.incremented():
+        for move_down, match, against in to_match:
+            if move_down:
+                with self.down_a_level():
                     self.match(match, against)
             else:
                 self.match(match, against)
@@ -319,25 +331,46 @@ class Matcher(object):
             names = get_match_names(match)
             if names:
                 tempvar = self.get_temp_var()
-                self.add_def(tempvar + " = "
-                             + "".join(kwargs + '.pop("' + name + '") if "' + name + '" in ' + kwargs + " else "
-                                       for name in names)
-                             + default
-                             )
-                with self.incremented():
+                self.add_def(
+                    tempvar + " = "
+                    + "".join(
+                        kwargs + '.pop("' + name + '") if "' + name + '" in ' + kwargs + " else "
+                        for name in names
+                    )
+                    + default,
+                )
+                with self.down_a_level():
                     self.match(match, tempvar)
             else:
                 raise CoconutDeferredSyntaxError("keyword-only pattern-matching function arguments must have names", self.loc)
 
     def match_dict(self, tokens, item):
         """Matches a dictionary."""
-        (matches,) = tokens
+        if len(tokens) == 1:
+            matches, rest = tokens[0], None
+        else:
+            matches, rest = tokens
         self.add_check("_coconut.isinstance(" + item + ", _coconut.abc.Mapping)")
-        self.add_check("_coconut.len(" + item + ") == " + str(len(matches)))
-        for x in range(len(matches)):
-            k, v = matches[x]
-            self.add_check(k + " in " + item)
-            self.match(v, item + "[" + k + "]")
+        if rest is None:
+            self.add_check("_coconut.len(" + item + ") == " + str(len(matches)))
+
+        if matches:
+            self.use_sentinel = True
+        for k, v in matches:
+            key_var = self.get_temp_var()
+            self.add_def(key_var + " = " + item + ".get(" + k + ", " + sentinel_var + ")")
+            with self.down_a_level():
+                self.add_check(key_var + " is not " + sentinel_var)
+                self.match(v, key_var)
+        if rest is not None and rest != wildcard:
+            match_keys = [k for k, v in matches]
+            with self.down_a_level():
+                self.add_def(
+                    rest + " = dict((k, v) for k, v in "
+                    + item + ".items() if k not in set(("
+                    + ", ".join(match_keys) + ("," if len(match_keys) == 1 else "")
+                    + ")))",
+                )
 
     def match_sequence(self, tokens, item):
         """Matches a sequence."""
@@ -351,35 +384,45 @@ class Matcher(object):
             self.add_check("_coconut.len(" + item + ") == " + str(len(matches)))
         else:
             self.add_check("_coconut.len(" + item + ") >= " + str(len(matches)))
-            if len(matches):
-                splice = "[" + str(len(matches)) + ":]"
-            else:
-                splice = ""
-            if series_type == "(":
-                self.add_def(tail + " = _coconut.tuple(" + item + splice + ")")
-            elif series_type == "[":
-                self.add_def(tail + " = _coconut.list(" + item + splice + ")")
-            else:
-                raise CoconutInternalException("invalid series match type", series_type)
+            if tail != wildcard:
+                if len(matches):
+                    splice = "[" + str(len(matches)) + ":]"
+                else:
+                    splice = ""
+                if series_type == "(":
+                    self.add_def(tail + " = _coconut.tuple(" + item + splice + ")")
+                elif series_type == "[":
+                    self.add_def(tail + " = _coconut.list(" + item + splice + ")")
+                else:
+                    raise CoconutInternalException("invalid series match type", series_type)
         self.match_all_in(matches, item)
 
     def match_iterator(self, tokens, item):
-        """Matches an iterator."""
+        """Matches a lazy list or a chain."""
         tail = None
         if len(tokens) == 2:
             _, matches = tokens
         else:
             _, matches, tail = tokens
         self.add_check("_coconut.isinstance(" + item + ", _coconut.abc.Iterable)")
-        itervar = self.get_temp_var()
         if tail is None:
+            itervar = self.get_temp_var()
             self.add_def(itervar + " = _coconut.tuple(" + item + ")")
-        else:
-            self.add_def(tail + " = _coconut.iter(" + item + ")")
+        elif matches:
+            itervar = self.get_temp_var()
+            if tail == wildcard:
+                tail = item
+            else:
+                self.add_def(tail + " = _coconut.iter(" + item + ")")
             self.add_def(itervar + " = _coconut.tuple(_coconut_igetitem(" + tail + ", _coconut.slice(None, " + str(len(matches)) + ")))")
-        with self.incremented():
-            self.add_check("_coconut.len(" + itervar + ") == " + str(len(matches)))
-            self.match_all_in(matches, itervar)
+        else:
+            itervar = None
+            if tail != wildcard:
+                self.add_def(tail + " = " + item)
+        if itervar is not None:
+            with self.down_a_level():
+                self.add_check("_coconut.len(" + itervar + ") == " + str(len(matches)))
+                self.match_all_in(matches, itervar)
 
     def match_star(self, tokens, item):
         """Matches starred assignment."""
@@ -395,16 +438,18 @@ class Matcher(object):
             head_matches, middle, last_matches = tokens
         self.add_check("_coconut.isinstance(" + item + ", _coconut.abc.Iterable)")
         if head_matches is None and last_matches is None:
-            self.add_def(middle + " = _coconut.list(" + item + ")")
+            if middle != wildcard:
+                self.add_def(middle + " = _coconut.list(" + item + ")")
         else:
             itervar = self.get_temp_var()
             self.add_def(itervar + " = _coconut.list(" + item + ")")
-            with self.incremented():
+            with self.down_a_level():
                 req_length = (len(head_matches) if head_matches is not None else 0) + (len(last_matches) if last_matches is not None else 0)
                 self.add_check("_coconut.len(" + itervar + ") >= " + str(req_length))
-                head_splice = str(len(head_matches)) if head_matches is not None else ""
-                last_splice = "-" + str(len(last_matches)) if last_matches is not None else ""
-                self.add_def(middle + " = " + itervar + "[" + head_splice + ":" + last_splice + "]")
+                if middle != wildcard:
+                    head_splice = str(len(head_matches)) if head_matches is not None else ""
+                    last_splice = "-" + str(len(last_matches)) if last_matches is not None else ""
+                    self.add_def(middle + " = " + itervar + "[" + head_splice + ":" + last_splice + "]")
                 if head_matches is not None:
                     self.match_all_in(head_matches, itervar)
                 if last_matches is not None:
@@ -416,45 +461,81 @@ class Matcher(object):
         front, series_type, matches = tokens
         self.add_check("_coconut.isinstance(" + item + ", _coconut.abc.Sequence)")
         self.add_check("_coconut.len(" + item + ") >= " + str(len(matches)))
-        if len(matches):
-            splice = "[:" + str(-len(matches)) + "]"
-        else:
-            splice = ""
-        if series_type == "(":
-            self.add_def(front + " = _coconut.tuple(" + item + splice + ")")
-        elif series_type == "[":
-            self.add_def(front + " = _coconut.list(" + item + splice + ")")
-        else:
-            raise CoconutInternalException("invalid series match type", series_type)
-        for x in range(len(matches)):
-            self.match(matches[x], item + "[" + str(x - len(matches)) + "]")
+        if front != wildcard:
+            if len(matches):
+                splice = "[:" + str(-len(matches)) + "]"
+            else:
+                splice = ""
+            if series_type == "(":
+                self.add_def(front + " = _coconut.tuple(" + item + splice + ")")
+            elif series_type == "[":
+                self.add_def(front + " = _coconut.list(" + item + splice + ")")
+            else:
+                raise CoconutInternalException("invalid series match type", series_type)
+        for i, match in enumerate(matches):
+            self.match(match, item + "[" + str(i - len(matches)) + "]")
 
     def match_msequence(self, tokens, item):
         """Matches a middle sequence."""
         series_type, head_matches, middle, _, last_matches = tokens
         self.add_check("_coconut.isinstance(" + item + ", _coconut.abc.Sequence)")
         self.add_check("_coconut.len(" + item + ") >= " + str(len(head_matches) + len(last_matches)))
-        if len(head_matches) and len(last_matches):
-            splice = "[" + str(len(head_matches)) + ":" + str(-len(last_matches)) + "]"
-        elif len(head_matches):
-            splice = "[" + str(len(head_matches)) + ":]"
-        elif len(last_matches):
-            splice = "[:" + str(-len(last_matches)) + "]"
-        else:
-            splice = ""
-        if series_type == "(":
-            self.add_def(middle + " = _coconut.tuple(" + item + splice + ")")
-        elif series_type == "[":
-            self.add_def(middle + " = _coconut.list(" + item + splice + ")")
-        else:
-            raise CoconutInternalException("invalid series match type", series_type)
+        if middle != wildcard:
+            if len(head_matches) and len(last_matches):
+                splice = "[" + str(len(head_matches)) + ":" + str(-len(last_matches)) + "]"
+            elif len(head_matches):
+                splice = "[" + str(len(head_matches)) + ":]"
+            elif len(last_matches):
+                splice = "[:" + str(-len(last_matches)) + "]"
+            else:
+                splice = ""
+            if series_type == "(":
+                self.add_def(middle + " = _coconut.tuple(" + item + splice + ")")
+            elif series_type == "[":
+                self.add_def(middle + " = _coconut.list(" + item + splice + ")")
+            else:
+                raise CoconutInternalException("invalid series match type", series_type)
         self.match_all_in(head_matches, item)
-        for x in range(len(last_matches)):
-            self.match(last_matches[x], item + "[" + str(x - len(last_matches)) + "]")
+        for i, match in enumerate(last_matches):
+            self.match(match, item + "[" + str(i - len(last_matches)) + "]")
+
+    def match_string(self, tokens, item):
+        """Match prefix string."""
+        prefix, name = tokens
+        return self.match_mstring((prefix, name, None), item, use_bytes=prefix.startswith("b"))
+
+    def match_rstring(self, tokens, item):
+        """Match suffix string."""
+        name, suffix = tokens
+        return self.match_mstring((None, name, suffix), item, use_bytes=suffix.startswith("b"))
+
+    def match_mstring(self, tokens, item, use_bytes=None):
+        """Match prefix and suffix string."""
+        prefix, name, suffix = tokens
+        if use_bytes is None:
+            if prefix.startswith("b") or suffix.startswith("b"):
+                if prefix.startswith("b") and suffix.startswith("b"):
+                    use_bytes = True
+                else:
+                    raise CoconutDeferredSyntaxError("string literals and byte literals cannot be added in patterns", self.loc)
+        if use_bytes:
+            self.add_check("_coconut.isinstance(" + item + ", _coconut.bytes)")
+        else:
+            self.add_check("_coconut.isinstance(" + item + ", _coconut.str)")
+        if prefix is not None:
+            self.add_check(item + ".startswith(" + prefix + ")")
+        if suffix is not None:
+            self.add_check(item + ".endswith(" + suffix + ")")
+        if name != wildcard:
+            self.add_def(
+                name + " = " + item + "[" +
+                ("" if prefix is None else "_coconut.len(" + prefix + ")") + ":"
+                + ("" if suffix is None else "-_coconut.len(" + suffix + ")") + "]",
+            )
 
     def match_const(self, tokens, item):
         """Matches a constant."""
-        (match,) = tokens
+        match, = tokens
         if match in const_vars:
             self.add_check(item + " is " + match)
         else:
@@ -462,7 +543,7 @@ class Matcher(object):
 
     def match_var(self, tokens, item):
         """Matches a variable."""
-        (setvar,) = tokens
+        setvar, = tokens
         if setvar != wildcard:
             if setvar in self.names:
                 self.add_check(self.names[setvar] + " == " + item)
@@ -472,10 +553,7 @@ class Matcher(object):
 
     def match_set(self, tokens, item):
         """Matches a set."""
-        if len(tokens) == 1:
-            match = tokens[0]
-        else:
-            raise CoconutInternalException("invalid set match tokens", tokens)
+        match, = tokens
         self.add_check("_coconut.isinstance(" + item + ", _coconut.abc.Set)")
         self.add_check("_coconut.len(" + item + ") == " + str(len(match)))
         for const in match:
@@ -501,28 +579,26 @@ class Matcher(object):
 
     def match_paren(self, tokens, item):
         """Matches a paren."""
-        (match,) = tokens
+        match, = tokens
         return self.match(match, item)
 
     def match_trailer(self, tokens, item):
         """Matches typedefs and as patterns."""
-        if len(tokens) <= 1 or len(tokens) % 2 != 1:
-            raise CoconutInternalException("invalid trailer match tokens", tokens)
-        else:
-            match, trailers = tokens[0], tokens[1:]
-            for i in range(0, len(trailers), 2):
-                op, arg = trailers[i], trailers[i + 1]
-                if op == "is":
-                    self.add_check("_coconut.isinstance(" + item + ", " + arg + ")")
-                elif op == "as":
-                    if arg in self.names:
-                        self.add_check(self.names[arg] + " == " + item)
-                    elif arg != wildcard:
-                        self.add_def(arg + " = " + item)
-                        self.names[arg] = item
-                else:
-                    raise CoconutInternalException("invalid trailer match operation", op)
-            self.match(match, item)
+        internal_assert(len(tokens) > 1 and len(tokens) % 2 == 1, "invalid trailer match tokens", tokens)
+        match, trailers = tokens[0], tokens[1:]
+        for i in range(0, len(trailers), 2):
+            op, arg = trailers[i], trailers[i + 1]
+            if op == "is":
+                self.add_check("_coconut.isinstance(" + item + ", " + arg + ")")
+            elif op == "as":
+                if arg in self.names:
+                    self.add_check(self.names[arg] + " == " + item)
+                elif arg != wildcard:
+                    self.add_def(arg + " = " + item)
+                    self.names[arg] = item
+            else:
+                raise CoconutInternalException("invalid trailer match operation", op)
+        self.match(match, item)
 
     def match_and(self, tokens, item):
         """Matches and."""
@@ -545,10 +621,12 @@ class Matcher(object):
 
     def out(self):
         out = ""
+        if self.use_sentinel:
+            out += sentinel_var + " = _coconut.object()\n"
         closes = 0
         for checks, defs in self.checkdefs:
             if checks:
-                out += "if (" + (") and (").join(checks) + "):\n" + openindent
+                out += "if " + paren_join(checks, "and") + ":\n" + openindent
                 closes += 1
             if defs:
                 out += "\n".join(defs) + "\n"
@@ -556,10 +634,11 @@ class Matcher(object):
             match_check_var + " = True\n"
             + closeindent * closes
             + "".join(other.out() for other in self.others)
-            + ("if " + match_check_var + " and not "
-                + ("(" + self.guards[0] + ")" if len(self.guards) == 1 else
-                    "((" + ") and (".join(self.guards) + "))")
-                + ":\n" + openindent
+            + (
+                "if " + match_check_var + " and not ("
+                + paren_join(self.guards, "and")
+                + "):\n" + openindent
                 + match_check_var + " = False\n" + closeindent
-                if self.guards else "")
+                if self.guards else ""
+            )
         )
