@@ -23,8 +23,8 @@ import sys as _coconut_sys
 # VERSION:
 #-----------------------------------------------------------------------------------------------------------------------
 
-VERSION = "1.2.3"
-VERSION_NAME = "Colonel"
+VERSION = "1.3.0"
+VERSION_NAME = "Dead Parrot"
 # False for release, int >= 1 for develop
 DEVELOP = False
 
@@ -34,7 +34,6 @@ DEVELOP = False
 
 if DEVELOP:
     VERSION += "-post_dev" + str(int(DEVELOP))
-__version__ = VERSION
 VERSION_STR = VERSION + " [" + VERSION_NAME + "]"
 
 PY2 = _coconut_sys.version_info < (3,)
@@ -53,8 +52,16 @@ class object(object):
         eq = self == other
         if eq is _coconut_NotImplemented:
             return eq
-        else:
-            return not eq
+        return not eq
+class int(_coconut_int):
+    __slots__ = ()
+    if hasattr(_coconut_int, "__doc__"):
+        __doc__ = _coconut_int.__doc__
+    class __metaclass__(type):
+        def __instancecheck__(cls, inst):
+            return _coconut.isinstance(inst, (_coconut_int, _coconut_long))
+        def __subclasscheck__(cls, subcls):
+            return _coconut.issubclass(subcls, (_coconut_int, _coconut_long))
 class range(object):
     __slots__ = ("_xrange",)
     if hasattr(_coconut_xrange, "__doc__"):
@@ -78,7 +85,7 @@ class range(object):
             return self._xrange[index]
     def count(self, elem):
         """Count the number of times elem appears in the range."""
-        return int(elem in self._xrange)
+        return _coconut_int(elem in self._xrange)
     def index(self, elem):
         """Find the index of elem in the range."""
         if elem not in self._xrange: raise _coconut.ValueError(_coconut.repr(elem) + " is not in range")
@@ -101,35 +108,31 @@ class range(object):
         return _coconut.isinstance(other, self.__class__) and self._args == other._args
 from collections import Sequence as _coconut_Sequence
 _coconut_Sequence.register(range)
-class int(_coconut_int):
-    __slots__ = ()
-    if hasattr(_coconut_int, "__doc__"):
-        __doc__ = _coconut_int.__doc__
-    class __metaclass__(type):
-        def __instancecheck__(cls, inst):
-            return _coconut.isinstance(inst, (_coconut_int, _coconut_long))
-        def __subclasscheck__(cls, subcls):
-            return _coconut.issubclass(subcls, (_coconut_int, _coconut_long))
 from functools import wraps as _coconut_wraps
 @_coconut_wraps(_coconut_print)
 def print(*args, **kwargs):
     file = kwargs.get("file", _coconut_sys.stdout)
+    flush = kwargs.get("flush", False)
+    if "flush" in kwargs:
+        del kwargs["flush"]
     if _coconut.hasattr(file, "encoding") and file.encoding is not None:
-        return _coconut_print(*(_coconut_unicode(x).encode(file.encoding) for x in args), **kwargs)
+        _coconut_print(*(_coconut_unicode(x).encode(file.encoding) for x in args), **kwargs)
     else:
-        return _coconut_print(*(_coconut_unicode(x).encode() for x in args), **kwargs)
+        _coconut_print(*(_coconut_unicode(x).encode() for x in args), **kwargs)
+    if flush:
+        file.flush()
 @_coconut_wraps(_coconut_raw_input)
 def input(*args, **kwargs):
     if _coconut.hasattr(_coconut_sys.stdout, "encoding") and _coconut_sys.stdout.encoding is not None:
         return _coconut_raw_input(*args, **kwargs).decode(_coconut_sys.stdout.encoding)
-    else:
-        return _coconut_raw_input(*args, **kwargs).decode()
+    return _coconut_raw_input(*args, **kwargs).decode()
 @_coconut_wraps(_coconut_repr)
 def repr(obj):
     if isinstance(obj, _coconut_unicode):
-        return _coconut_repr(obj)[1:]
-    else:
-        return _coconut_repr(obj)
+        return _coconut_unicode(_coconut_repr(obj)[1:])
+    if isinstance(obj, _coconut_str):
+        return "b" + _coconut_unicode(_coconut_repr(obj))
+    return _coconut_unicode(_coconut_repr(obj))
 ascii = repr
 def raw_input(*args):
     """Coconut uses Python 3 "input" instead of Python 2 "raw_input"."""
